@@ -24,16 +24,19 @@
 
 from __future__ import absolute_import, division, print_function, unicode_literals
 
-import datetime, errno, json, math, os, re, select, shlex, sys, time
+import datetime, errno, json, math, numbers, os, re, select, shlex, sys, time
 from subprocess import check_output, Popen, PIPE
 from xml.dom import minidom
 
-from .. import filesystem
-from ..datetime import datetime_now, parts_to_time, secs_to_time, str2time, time_ratio
+from .. import comparison, filesystem, validation
+from ..datetime import datetime_now, parts_to_time, secs_to_time, str_to_time, time_ratio
 from ..encoding import string_types
 from ..subprocess import make_async
 
-__all__ = ('ENCODING_REGEX', 'DURATION_REGEX', 'WIDTH', 'HEIGHT', 'MPD_TEST', 'FFmpeg')
+__all__ = (
+    'ENCODING_REGEX', 'DURATION_REGEX', 'WIDTH', 'HEIGHT', 'MPD_TEST', 'Codec', 'Stream', 'AudioStream', 'VideoStream',
+    'Media', 'FFmpeg'
+)
 
 # frame= 2071 fps=  0 q=-1.0 size=   34623kB time=00:01:25.89 bitrate=3302.3kbits/s
 ENCODING_REGEX = re.compile(
@@ -50,12 +53,117 @@ MPD_TEST = """<?xml version="1.0"?>
 """
 
 
+def _to_args_list(value):
+    return (shlex.split(value) if isinstance(value, string_types) else value) or []
+
+
+def _to_framerate(fps):
+    """
+    Return the frame rate as float or None in case of error.
+
+    **Example usage**
+
+    >>> print(_to_framerate({}))
+    None
+    >>> print(_to_framerate(25.0))
+    25.0
+    >>> print(_to_framerate('59000/1000'))
+    59.0
+    """
+    try:
+        if isinstance(fps, numbers.Number):
+            return fps
+        if '/' in fps:
+            num, denom = fps.split('/')
+            return float(num) / float(denom)
+        return float(fps)
+    except:
+        return None
+
+
+class Codec(validation.CleanAttributesMixin, comparison.SlotsEqualityMixin):
+
+    __slots__ = ('long_name', 'name', 'tag', 'tag_string', 'time_base', 'type')
+
+    def __init__(self, infos):
+        for attr in self.__slots__:
+            setattr(self, attr, infos['codec_' + attr])
+
+    clean_time_base = lambda s, v: _to_framerate(v)
+
+
+class Stream(validation.CleanAttributesMixin, comparison.SlotsEqualityMixin):
+
+    codec_class = Codec
+    defaults = {}
+
+    def __init__(self, infos):
+        for attr in self.__slots__:
+            if attr == 'codec':
+                self.codec = self.codec_class(infos)
+            else:
+                self._set_attribute(attr, infos)
+
+    def _set_attribute(self, name, infos):
+        """Set attribute ``name`` value from the ``infos`` or ``self.defaults`` dictionary."""
+        setattr(self, name, infos.get(name, self.defaults.get(name)))
+
+
+class AudioStream(Stream):
+
+    __slots__ = (
+        'avg_frame_rate', 'bit_rate', 'bits_per_sample', 'channel_layout', 'channels', 'codec', 'disposition',
+        'duration', 'duration_ts', 'index', 'nb_frames', 'r_frame_rate', 'sample_fmt', 'sample_rate', 'start_pts',
+        'start_time', 'tags', 'time_base'
+    )
+
+    clean_avg_frame_rate = clean_r_frame_rate = clean_time_base = lambda s, v: None if v is None else _to_framerate(v)
+    clean_bit_rate = clean_bits_per_sample = clean_channels = clean_duration_ts = clean_index = clean_nb_frames = \
+        clean_sample_rate = clean_start_pts = lambda s, v: None if v is None else int(v)
+    clean_duration = clean_start_time = lambda s, v: None if v is None else float(v)
+
+
+class VideoStream(Stream):
+
+    __slots__ = (
+        'avg_frame_rate', 'codec', 'display_aspect_ratio', 'disposition', 'has_b_frames', 'height', 'index', 'level',
+        'pix_fmt', 'r_frame_rate', 'sample_aspect_ratio', 'time_base', 'width'
+    )
+
+    clean_avg_frame_rate = clean_r_frame_rate = clean_time_base = lambda s, v: None if v is None else _to_framerate(v)
+    clean_height = clean_index = clean_level = clean_width = lambda s, v: None if v is None else int(v)
+
+
+class Media(validation.CleanAttributesMixin, comparison.SlotsEqualityMixin):
+
+    __slots__ = ('filename', 'options')
+
+    def __init__(self, filename, options=None):
+        self.filename = filename
+        self.options = options
+
+    @property
+    def directory(self):
+        return os.path.abspath(os.path.dirname(self.filename))
+
+    def clean_options(self, value):
+        return _to_args_list(value)
+
+    def create_directory(self):
+        filesystem.try_makedirs(self.directory)
+
+    def to_args(self, is_input):
+        return self.options + (['-i', self.filename] if is_input else [self.filename])
+
+
 class FFmpeg(object):
 
     duration_regex = DURATION_REGEX
     encoding_regex = ENCODING_REGEX
     encoding_executable = 'ffmpeg'
     parsing_executable = 'ffprobe'
+    media_class = Media
+    stream_classes = {'audio': None, 'video': None}
 
     def __init__(self, encoding_executable=None, parsing_executable=None,
                  default_in_duration=datetime.timedelta(seconds=0), ratio_delta=0.01, time_delta=1, max_time_delta=5,
@@ -70,45 +178,72 @@ class FFmpeg(object):
         self.sanity_max_ratio = sanity_max_ratio
         self.encoding = encoding
 
-    def _get_arguments(self, in_filenames, out_filename, options):
+    def _clean_medias_argument(self, value):
         """
-        Return the arguments for the encoding process.
-
-        * Set in_filenames to a string or a "list" with the input filenames.
-        * Set out_filename to a string.
-        * Set argument to a string or a list with the options for the process (except ones related to input(s)/output).
-
-        In return you will get a tuple with (arguments, in_filenames -> list, out_filename -> str, options -> list).
+        Return a list of Media instances from passed value. Value can be one or multiple instances of string or Media.
 
         **Example usage**
 
         >>> from nose.tools import assert_list_equal as leq_
-        >>> options_string = '-strict experimental -vf "yadif=0.-1:0, scale=trunc(iw/2)*2:trunc(ih/2)*2"'
-        >>> get_args = FFmpeg()._get_arguments
+        >>> handle = FFmpeg()._clean_medias_argument
 
-        >>> args, in_filenames, out_filename, options = get_args('input.mp4', 'output.mkv', options_string)
-        >>> leq_(in_filenames, ['input.mp4'])
+        >>> leq_(handle(None), [])
+        >>> leq_(handle([]), [])
+        >>> leq_(handle('a.mp4'), [Media('a.mp4')])
+        >>> leq_(handle(['a.mp4', 'b.mp3']), [Media('a.mp4'), Media('b.mp3')])
+        >>> leq_(handle(Media('a', '-f mp4')), [Media('a', ['-f', 'mp4'])])
+        >>> leq_(handle([Media('a', ['-f', 'mp4']), Media('b.mp3')]), [Media('a', ['-f', 'mp4']), Media('b.mp3')])
+        """
+        values = [value] if isinstance(value, (string_types, Media)) else value
+        return [self.media_class(v) if isinstance(v, string_types) else v for v in values] if values else []
+
+    def _get_arguments(self, inputs, outputs, options=None):
+        """
+        Return the arguments for the encoding process.
+
+        * Set inputs to one or multiple strings (filenames) or Media instances (with options).
+        * Set outputs to one or multiple strings (filenames) or Media instances (with options).
+        * Set options to a string or a list with the options to put in-between the inputs and outputs (legacy API).
+
+        In return you will get a tuple with (arguments, inputs -> list Media, outputs -> list Media, options -> list).
+
+        **Example usage**
+
+        >>> from nose.tools import assert_list_equal as leq_
+        >>> get = FFmpeg()._get_arguments
+
+        Using options (the legacy API, also simplify simple calls):
+
+        >>> options_string = '-strict experimental -vf "yadif=0.-1:0, scale=trunc(iw/2)*2:trunc(ih/2)*2"'
+
+        >>> args, inputs, outputs, options = get('input.mp4', 'output.mkv', options_string)
+        >>> leq_(inputs, [Media('input.mp4')])
+        >>> leq_(outputs, [Media('output.mkv')])
         >>> leq_(options, ['-strict', 'experimental', '-vf', 'yadif=0.-1:0, scale=trunc(iw/2)*2:trunc(ih/2)*2'])
         >>> leq_(args, ['ffmpeg', '-y', '-i', 'input.mp4'] + options + ['output.mkv'])
 
-        >>> args, in_filenames, out_filename, options = get_args({'input.avi'}, 'output.mp4', options_string)
-        >>> leq_(in_filenames, ['input.avi'])
-        >>> leq_(options, ['-strict', 'experimental', '-vf', 'yadif=0.-1:0, scale=trunc(iw/2)*2:trunc(ih/2)*2'])
-        >>> leq_(args, ['ffmpeg', '-y', '-i', 'input.avi'] + options + ['output.mp4'])
+        >>> args, _, outputs, options = get('input.mp4', None, options_string)
+        >>> leq_(outputs, [])
+        >>> leq_(args, ['ffmpeg', '-y', '-i', 'input.mp4'] + options)
 
-        >>> args, in_filenames, out_filename, options = get_args(('video.h264', 'audio.mp3'), 'output.mp4', options)
-        >>> leq_(sorted(in_filenames), ['audio.mp3', 'video.h264'])
-        >>> leq_(options, ['-strict', 'experimental', '-vf', 'yadif=0.-1:0, scale=trunc(iw/2)*2:trunc(ih/2)*2'])
-        >>> leq_(args, ['ffmpeg', '-y', '-i', 'video.h264', '-i', 'audio.mp3'] + options + ['output.mp4'])
+        Using instances of Media (the newest API, greater flexibility):
+
+        >>> args, inputs, outputs, options = get(Media('in', '-f mp4'), Media('out.mkv', '-acodec copy -vcodec copy'))
+        >>> leq_(inputs, [Media('in', ['-f', 'mp4'])])
+        >>> leq_(outputs, [Media('out.mkv', ['-acodec', 'copy', '-vcodec', 'copy'])])
+        >>> leq_(options, [])
+        >>> leq_(args, ['ffmpeg', '-y', '-f', 'mp4', '-i', 'in', '-acodec', 'copy', '-vcodec', 'copy', 'out.mkv'])
         """
-        in_filenames = [f for f in ([in_filenames] if isinstance(in_filenames, string_types) else in_filenames)]
-        options = (shlex.split(options) if isinstance(options, string_types) else options) or []
-
+        inputs = self._clean_medias_argument(inputs)
+        outputs = self._clean_medias_argument(outputs)
+        options = _to_args_list(options)
         args = [self.encoding_executable, '-y']
-        for in_filename in in_filenames:
-            args.extend(['-i', in_filename])
-        args.extend(options + [out_filename])
-        return args, in_filenames, out_filename, options
+        for the_input in inputs:
+            args.extend(the_input.to_args(is_input=True))
+        args.extend(options)
+        for output in outputs:
+            args.extend(output.to_args(is_input=False))
+        return args, inputs, outputs, options
 
     def _get_chunk(self, process):
         select.select([process.stderr], [], [])
@@ -135,7 +270,7 @@ class FFmpeg(object):
         return process
 
     def _get_progress(self, in_duration, stats):
-        out_duration = str2time(stats['time'])
+        out_duration = str_to_time(stats['time'])
         ratio = time_ratio(out_duration, in_duration)
         return out_duration, ratio
 
@@ -215,38 +350,6 @@ class FFmpeg(object):
                 return duration
         return None
 
-    def get_media_framerate(self, filename_or_infos):
-        """
-        Return the frame rate of the first video stream in ``filename_or_infos`` or None in case of error.
-
-        **Example usage**
-
-        >>> print(FFmpeg().get_media_framerate(3.14159265358979323846))
-        None
-        >>> print(FFmpeg().get_media_framerate({}))
-        None
-        >>> print(FFmpeg().get_media_framerate(FFmpeg().get_media_infos('small.mp4')))
-        30.0
-        >>> print(FFmpeg().get_media_framerate('small.mp4'))
-        30.0
-        >>> print(FFmpeg().get_media_framerate({'streams': [
-        ...     {'codec_type': 'audio'},
-        ...     {'codec_type': 'video', 'avg_frame_rate': '59000/1000'}
-        ... ]}))
-        59.0
-        """
-        if not isinstance(filename_or_infos, dict):
-            filename_or_infos = self.get_media_infos(filename_or_infos)
-        try:
-            first_video_stream = next(s for s in filename_or_infos['streams'] if s['codec_type'] == 'video')
-            fps = first_video_stream['avg_frame_rate']
-            if '/' in fps:
-                num, denom = fps.split('/')
-                return float(num) / float(denom)
-            return float(fps)
-        except:
-            return None
-
     def get_media_infos(self, filename):
         """
         Return a Python dictionary containing informations about the media informations or None in case of error.
@@ -354,34 +457,123 @@ class FFmpeg(object):
         except:
             return None
 
-    def get_media_resolution(self, filename_or_infos):
+    def get_streams(self, filename_or_infos, condition=lambda stream: True):
+        infos = filename_or_infos if isinstance(filename_or_infos, dict) else self.get_media_infos(filename_or_infos)
+        try:
+            raw_streams = (s for s in infos['streams'] if condition(s))
+        except:
+            return []
+        streams = []
+        for stream in raw_streams:
+            stream_class = self.stream_classes[stream['codec_type']]
+            streams.append(stream_class(stream) if stream_class and not isinstance(stream, stream_class) else stream)
+        return streams
+
+    def get_audio_streams(self, filename_or_infos):
         """
-        Return [width, height] of the first video stream in ``filename_or_infos`` or None in case of error.
+        Return a list with the audio streams ``filename_or_infos`` or [] in case of error.
 
         **Example usage**
 
-        >>> print(FFmpeg().get_media_resolution(3.14159265358979323846))
+        >>> from nose.tools import eq_
+        >>> ffmpeg = FFmpeg()
+
+        >>> ffmpeg.stream_classes['audio'] = None
+        >>> streams = ffmpeg.get_audio_streams('small.mp4')
+        >>> assert isinstance(streams[0], dict)
+        >>> eq_(streams[0]['avg_frame_rate'], '0/0')
+        >>> eq_(streams[0]['channels'], 1)
+        >>> eq_(streams[0]['codec_time_base'], '1/48000')
+
+        >>> ffmpeg.stream_classes['audio'] = AudioStream
+        >>> streams = ffmpeg.get_audio_streams('small.mp4')
+        >>> assert isinstance(streams[0], AudioStream)
+        >>> eq_(streams[0].avg_frame_rate, None)
+        >>> eq_(streams[0].channels, 1)
+        >>> eq_(streams[0].codec.time_base, 1 / 48000)
+        """
+        return self.get_streams(filename_or_infos, condition=lambda s: s['codec_type'] == 'audio')
+
+    def get_video_streams(self, filename_or_infos):
+        """
+        Return a list with the video streams ``filename_or_infos`` or [] in case of error.
+
+        **Example usage**
+
+        >>> from nose.tools import eq_
+        >>> ffmpeg = FFmpeg()
+
+        >>> ffmpeg.stream_classes['video'] = None
+        >>> streams = ffmpeg.get_video_streams('small.mp4')
+        >>> assert isinstance(streams[0], dict)
+        >>> eq_(streams[0]['avg_frame_rate'], '30/1')
+
+        >>> ffmpeg.stream_classes['video'] = VideoStream
+        >>> streams = ffmpeg.get_video_streams('small.mp4')
+        >>> assert isinstance(streams[0], VideoStream)
+        >>> eq_(streams[0].avg_frame_rate, 30.0)
+        """
+        return self.get_streams(filename_or_infos, condition=lambda s: s['codec_type'] == 'video')
+
+    def get_video_framerate(self, filename_or_infos, index=0, fail=False):
+        """
+        Return the frame rate of the video stream at ``index`` in ``filename_or_infos`` or None in case of error.
+
+        **Example usage**
+
+        >>> print(FFmpeg().get_video_framerate(3.14159265358979323846))
         None
-        >>> print(FFmpeg().get_media_resolution({}))
+        >>> print(FFmpeg().get_video_framerate({}))
         None
-        >>> print(FFmpeg().get_media_resolution(FFmpeg().get_media_infos('small.mp4')))
+        >>> print(FFmpeg().get_video_framerate(FFmpeg().get_media_infos('small.mp4')))
+        30.0
+        >>> print(FFmpeg().get_video_framerate('small.mp4'))
+        30.0
+        >>> print(FFmpeg().get_video_framerate({'streams': [
+        ...     {'codec_type': 'audio'},
+        ...     {'codec_type': 'video', 'avg_frame_rate': '59000/1000'}
+        ... ]}))
+        59.0
+        """
+        try:
+            stream = self.get_video_streams(filename_or_infos)[index]
+            return _to_framerate(stream['avg_frame_rate']) if isinstance(stream, dict) else stream.avg_frame_rate
+        except:
+            if fail:
+                raise
+            return None
+
+    def get_video_resolution(self, filename_or_infos, index=0, fail=False):
+        """
+        Return [width, height] of the video stream at ``index`` in ``filename_or_infos`` or None in case of error.
+
+        **Example usage**
+
+        >>> print(FFmpeg().get_video_resolution(3.14159265358979323846))
+        None
+        >>> print(FFmpeg().get_video_resolution({}))
+        None
+        >>> print(FFmpeg().get_video_resolution(FFmpeg().get_media_infos('small.mp4')))
         [560, 320]
-        >>> print(FFmpeg().get_media_resolution('small.mp4'))
+        >>> print(FFmpeg().get_video_resolution('small.mp4'))
         [560, 320]
-        >>> print(FFmpeg().get_media_resolution('small.mp4')[HEIGHT])
+        >>> print(FFmpeg().get_video_resolution('small.mp4', index=1))
+        None
+        >>> print(FFmpeg().get_video_resolution('small.mp4')[HEIGHT])
         320
-        >>> print(FFmpeg().get_media_resolution({'streams': [
+        >>> print(FFmpeg().get_video_resolution({'streams': [
         ...     {'codec_type': 'audio'},
         ...     {'codec_type': 'video', 'width': '1920', 'height': '1080'}
         ... ]}))
         [1920, 1080]
         """
-        if not isinstance(filename_or_infos, dict):
-            filename_or_infos = self.get_media_infos(filename_or_infos)
         try:
-            first_video_stream = next(s for s in filename_or_infos['streams'] if s['codec_type'] == 'video')
-            return [int(first_video_stream['width']), int(first_video_stream['height'])]
+            stream = self.get_video_streams(filename_or_infos)[index]
+            is_dict = isinstance(stream, dict)
+            return [int(stream['width']), int(stream['height'])] if is_dict else [stream.width, stream.height]
         except:
+            if fail:
+                raise
             return None
 
     def get_now(self):
@@ -390,9 +582,11 @@ class FFmpeg(object):
     def get_size(self, path):
         return filesystem.get_size(path)
 
-    def encode(self, in_filenames, out_filename, options, base_track=0, create_out_directory=True):
+    def encode(self, inputs, outputs, options=None, base_track=0, create_directories=True):
         """
-        Encode a set of input files input an output file.
+        Encode a set of input files input to a set of output files.
+
+        .. note:: Current implementation only handles a unique output.
 
         **Example usage**
 
@@ -400,37 +594,38 @@ class FFmpeg(object):
 
         >>> encoder = FFmpeg()
 
-        >>> results = list(encoder.encode('small.mp4', 'ff_output.mp4', '-c:a copy -c:v copy'))
+        >>> results = list(encoder.encode(Media('small.mp4'), Media('ff_output.mp4', '-c:a copy -c:v copy')))
         >>> try_remove('ff_output.mp4')
         True
         >>> print(results[-1]['status'])
         SUCCESS
 
-        >>> results = list(encoder.encode('small.mp4', 'ff_output.mp4', 'crazy_option'))
+        >>> results = list(encoder.encode(Media('small.mp4'), Media('ff_output.mp4', 'crazy_option')))
         >>> try_remove('ff_output.mp4')
         False
         >>> print(results[-1]['status'])
         ERROR
 
-        >>> results = list(encoder.encode({'missing.mp4'}, 'ff_output.mp4', '-c:a copy -c:v copy'))
+        >>> results = list(encoder.encode([Media('missing.mp4')], Media('ff_output.mp4', '-c:a copy -c:v copy')))
         >>> try_remove('ff_output.mp4')
         False
         >>> print(results[-1]['status'])
         ERROR
         """
-        arguments, in_filenames, out_filename, options = self._get_arguments(in_filenames, out_filename, options)
+        arguments, inputs, outputs, _ = self._get_arguments(inputs, outputs, options)
+        if len(outputs) != 1:
+            raise NotImplementedError()
 
-        # Create output directory
-        if create_out_directory:
-            directory = os.path.dirname(out_filename)
-            if directory:
-                filesystem.try_makedirs(directory)
+        # Create outputs directories
+        if create_directories:
+            for output in outputs:
+                output.create_directory()
 
         process = self._get_process(arguments)
         try:
             # Get input media duration and size to be able to estimate ETA
-            in_duration = self.get_media_duration(in_filenames[base_track]) or self.default_in_duration
-            in_size = self.get_size(in_filenames[base_track])
+            in_duration = self.get_media_duration(inputs[base_track].filename) or self.default_in_duration
+            in_size = self.get_size(inputs[base_track].filename)
 
             # Initialize metrics
             output = ''
@@ -458,16 +653,16 @@ class FFmpeg(object):
                         yield self._clean_statistics(
                             stats=stats, status='PROGRESS', output=output, returncode=None, start_date=start_date,
                             elapsed_time=datetime.timedelta(seconds=elapsed_time), ratio=ratio, in_size=in_size,
-                            in_duration=in_duration, out_size=self.get_size(out_filename), out_duration=out_duration,
-                            frame=int(stats['frame']), fps=float(stats['fps']), bitrate=stats['bitrate'],
-                            quality=stats.get('q'), sanity=None
+                            in_duration=in_duration, out_size=self.get_size(outputs[0].filename),
+                            out_duration=out_duration, frame=int(stats['frame']), fps=float(stats['fps']),
+                            bitrate=stats['bitrate'], quality=stats.get('q'), sanity=None
                         )
                 returncode = process.poll()
                 if returncode is not None:
                     break
 
             # Output media file sanity check
-            out_duration = self.get_media_duration(out_filename)
+            out_duration = self.get_media_duration(outputs[0].filename)
             ratio = time_ratio(out_duration, in_duration) if out_duration else 0.0
             frame = int(stats.get('frame', 0))  # FIXME compute latest frame based on output infos
             fps = float(stats.get('fps', 0))  # FIXME compute average fps
@@ -475,8 +670,8 @@ class FFmpeg(object):
                 stats=stats, status='ERROR' if returncode else 'SUCCESS', output=output, returncode=returncode,
                 start_date=start_date, elapsed_time=datetime.timedelta(seconds=elapsed_time),
                 eta_time=datetime.timedelta(0), ratio=ratio if returncode else 1.0, in_size=in_size,
-                in_duration=in_duration, out_size=self.get_size(out_filename), out_duration=out_duration, frame=frame,
-                fps=fps, bitrate=stats.get('bitrate'), quality=stats.get('q'),
+                in_duration=in_duration, out_size=self.get_size(outputs[0].filename), out_duration=out_duration,
+                frame=frame, fps=fps, bitrate=stats.get('bitrate'), quality=stats.get('q'),
                 sanity=self.sanity_min_ratio <= ratio <= self.sanity_max_ratio
             )
         except Exception as exception:
